# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_items.ipynb (unless otherwise specified).

__all__ = ['Item', 'Collection', 'Selection', 'ItemList', 'ItemArray']

# Cell
from traitlets import HasTraits, Tuple, Unicode, Any, observe, List, Int, Float
import traitlets
import numpy as np
from traittypes import Array
from collections import namedtuple
import warnings

# Cell
class Item(HasTraits):
    '''The <code>name</code> property is used by model collections and view collections
    in order to store and retrieve models and views by name'''
    def __init__(self, name=None, **kwargs):
        super().__init__(**kwargs)
        if name:
            self.name = name
        self._index = {}

    def _repr_pretty_(self, p, cycle):
        if hasattr(self, 'name'):
            name = self.name
        else:
            name = self.__class__.__name__
        with p.group(4, str(name) +'(', ')'):
            p.breakable()
            for trait, value in self._trait_values.items():
                p.text(trait + ': ')
                p.pretty(value)
                p.text(',')
                p.breakable()

# Cell
class Collection(Item):
    '''Supports 2D and 3D arrays. 3D arrays can be accessed with tuples but are stored in
    a list.'''

    children = Tuple()

    def __init__(self, children=None, **kwargs):
        super().__init__(**kwargs)
        if children:
            self.children += tuple(children)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._children[key]
        elif isinstance(key, slice):
            return [self._children[i] for i in range(*key.indices(self._size))]
        else:
            raise KeyError('Key must be an int or slice')

    def __iter__(self):
        return self._children.__iter__()

    def get_child_index(self):
        return self._size

    @observe('children')
    def _observe_collection_children(self, change):
        self._children = ()
        self._size = 0
        for child in change['new']:
            self._add_child(child)
        return tuple(self._children)

    #def add_collection(self, collection):
    #    self._collections[id(collection)] = collection
    #    self.children += collection.children

    def _add_child(self, child):
        '''Modify child and add child to list self._children'''
        child._index[id(self)] = self.get_child_index()
        #self._indices[id(child)] = child._index[id(self)]
        self._size += 1
        self._children += (child, )

    def get_links(self, attr):
        return [(child, attr) for child in self._children]

# Cell
class Selection(Item):
    '''A collection with selectable item.'''

    children = Tuple()
    _children = Tuple()
    selected_index = Int(default_value=None, allow_none=True)
    selected_name = Unicode(default_value=None, allow_none=True)

    def __init__(self, children, **kwargs):
        super().__init__(**kwargs)
        self._size = 0
        self._children = children

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._children[key]
        elif isinstance(key, slice):
            return [self._children[i] for i in range(*key._indices(self._size))]
        elif isinstance(key, str):
            if key in self._indices:
                return self._children[self._indices[key]]
            else:
                raise KeyError('Key %s not found in children or collections' % str(key))

    def __iter__(self):
        return self._children.__iter__()

    #@validate('children')
    #TODO: children can only be changed to match selected name and index

    def add_child(self, child, name=None):
        if name:
            child.name = name
        self._children += (child, )

    @observe('_children')
    def _observe_selection_children(self, change):
        self._names = []
        self._indices = {} # name: index
        ret = tuple([self._add_child(child, i) for i, child in enumerate(change['new'])])
        if ret:
            self.selected_index = 0
            self.selected_name = ret[0].name
            self.children = (ret[0], )
        self._size = len(ret)
        return ret

    def _add_child(self, child, index):
        '''Modify child and add child to list self._children'''
        child.item_index = index
        if not hasattr(child, 'name'):
            child.name = child.__class__.__name__ + str(self._size)
        self._names.append(child.name)
        self._indices[child.name] = child.item_index
        return child

    @observe('selected_index')
    def _observe_selected_index(self, change):
        selected_index = change['new']
        if self._size:
            self.children = (self._children[selected_index], )
            self.selected_name = self._names[selected_index]
        return selected_index

    @observe('selected_name')
    def _observe_selected_name(self, change):
        selected_name = change['new']
        if self._size:
            self.selected_index = self._indices[selected_name]
        return selected_name

# Cell
class ItemList(Collection):
    '''A WidgetCollectin in which every child is the same type of Widget.
    Any child trait name listed in `trait_names` will be turned into an
    observable list, for example if 'value' is listed, a List trait 'value_list'
    will give the user access to a list of child values so that when the list
    is updated, so is the value of the child'''

    def __init__(self, children, traits=('value', ), **kwargs):
        super().__init__(**kwargs)
        self._traits = traits
        self.children = tuple(children)

    @observe('children')
    def _observe_collection_children(self, change):
        super()._observe_collection_children(change)
        if self._children and self._traits:
            for trait in self._traits:
                # initialize and observe child handler
                for child in self._children:
                    child._child_handlers[trait] = traitlets.observe(trait, type='change')
                    child._child_handlers[trait]._init_call(self._child_handler_factory(trait))
                    child.observe(child._child_handlers[trait], trait)

    def _add_child(self, child):
        child._child_handlers = {}
        super()._add_child(child)

    def set_list(self, values, trait='value'):
        if isinstance(values, np.ndarray):
            values = values.tolist()
        for i, value in enumerate(values):
            setattr(self._children[i], trait, value)

    def get_list(self, values, trait='value'):
        ret = []
        for i, child in enumerate(self._children):
            ret.append(getattr(child, trait))
        return ret

    def apply_all(self, method_name, *args, **kwargs):
        for child in self._children:
            method = getattr(child, method_name)
            method(*args, **kwargs)

    def observe_all(self, handler, names=None, type='change'):
        for child in self._children:
            if names:
                child.observe(handler, names, type)
            else:
                child.observe(handler)

    def unobserve_all(self, handler, names=None, type='change'):
        for child in self._children:
            if names:
                child.unobserve(handler, names, type)
            else:
                child.unobserve(handler)

    def observe_children(self, trait):
        for child in self._children:
            child.observe(child._child_handlers[trait], trait)

    def unobserve_children(self, trait):
        for child in self._children:
            child.unobserve(child._child_handlers[trait], trait)

    def _child_handler_factory(self, name):
        def wrapper(change):
            return self._child_handler(name, change)
        return wrapper

    def _child_handler(self, name, change):
        # add to change dict and notify change
        change['index'] = change['owner']._index[id(self)]
        change['type'] = 'child_change'
        change['list'] = self.get_list(name)
        self.notify_change(change)
        return change['new']

# Cell
class ItemArray(ItemList):
    '''A CollectionList with two dimensions. Children can be accessed
    with tuple notation similar to the numpy library. Any trait_name listed
    will cause the creation of an array of child traits, for example
    if 'value' is listed, the trait 'value_array' will become available'''
    def __init__(self, n_rows, n_columns, children=None, traits='values', **kwargs):
        self.n_rows = n_rows
        self.n_columns = n_columns
        super().__init__(children, traits, **kwargs)

    def get_child_index(self):
        return self._2D_index(self._size)

    def set_array(self, values, trait='value'):
        #self.unobserve_children(trait)
        if isinstance(values, np.ndarray):
            values = values.tolist()
        for i in range(self.n_rows):
            for j in range(self.n_columns):
                setattr(self._children[self._flat_index(i, j)], trait, values[i][j])
        #self.observe_children(trait)

    def get_array(self, trait='value'):
        ret = []
        for i in range(self.n_rows):
            ret.append([])
            for j in range(self.n_columns):
                ret[i].append(getattr(self._children[self._flat_index(i, j)], trait))
        return ret

    def __getitem__(self, index):
        if not isinstance(index, tuple):
            raise IndexError('Array index must be a tuple')
        x, y = index
        if isinstance(x, int) and isinstance(y, int):
            assert x <= self.n_rows and y <= self.n_columns, "index (%i, %i) not in range" % (i, j)
            return self._children[self._flat_index(x, y)]
        else:
            ret = []
            for x in self._to_range(x, self.n_rows):
                for y in self._to_range(y, self.n_columns):
                    ret.append(self._children[self._flat_index(x, y)])
            return ret

    def _to_range(self, s, end):
        if isinstance(s, int):
            return range(s, s+1)
        else:
            if hasattr(s, 'stop'):
                assert s.stop <= end, "slice not in range"
                end = s.stop
            return range(*s.indices(end))

    def _flat_index(self, i, j):
        return i * self.n_columns + j

    def _2D_index(self, index):
        row = int(index / self.n_columns)
        col = int(index % self.n_columns)
        return (row, col)

    def _child_handler(self, name, change):
        # add to change dict and notify change
        change['index'] = change['owner']._index[id(self)]
        change['type'] = 'child_change'
        change['array'] = self.get_array(name)
        self.notify_change(change)
        return change['new']


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_config.ipynb (unless otherwise specified).

__all__ = ['APP_DIR', 'CONFIG', 'USER_DIR']

# Cell
import os, pathlib, urllib, configparser
import pprint

# Cell
APP_DIR = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))

# Cell
CONFIG = configparser.ConfigParser(allow_no_value=True)
CONFIG.read(os.path.join(APP_DIR, 'default.ini'))

# Cell
USER_DIR = os.path.join(os.path.expanduser("~"),'.superpower_gui')
if not os.path.exists(USER_DIR):
    os.makedirs(USER_DIR)
    with open(os.path.join(USER_DIR, 'config.ini'), 'w') as configfile:
        configfile.write('# This config file was created to match default app settings. You may want to change them.\n')
        CONFIG.write(configfile)

# Cell
if os.path.isfile(os.path.join(USER_DIR, 'config.ini')):
    CONFIG.read(os.path.join(USER_DIR, 'config.ini'))

if CONFIG['LOG']['filename']:
    CONFIG['LOG']['filename'] = os.path.expanduser(CONFIG['LOG']['filename'])
    if not os.path.isabs(CONFIG['LOG']['filename']):
        CONFIG['LOG']['filename'] = os.path.join(USER_DIR, CONFIG['LOG']['filename'])
    assert os.path.isdir(os.path.dirname(CONFIG['LOG']['filename'])), 'Directory %s doesn\'t exist' % CONFIG['LOG']['filename']


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_widget_containers.ipynb (unless otherwise specified).

__all__ = ['WidgetContainer', 'Box', 'HBox', 'VBox', 'WidgetList', 'WidgetRow', 'WidgetColumn', 'SelectionContainer',
           'SelectionBox', 'SelectionVBox', 'SelectionHBox', 'TabAcc', 'Tab', 'Accordion']

# Cell
from collections import namedtuple
import traitlets
from traitlets import Type, Unicode, observe, validate, Any, HasTraits, directional_link, link, Tuple, List, Instance
import ipywidgets as ipyw
import numpy as np

# Cell
from .items import Collection, Selection, ItemList, ItemArray
from .widgets import Widget, FloatText, IntText, Label, Grid, FloatSlider

# Cell
class WidgetContainer(Collection, Widget):

    children = ipyw.trait_types.TypedTuple(help="List of widget children").tag(sync=True, **ipyw.widget.widget_serialization)

class Box(WidgetContainer, ipyw.Box):
    pass

class HBox(WidgetContainer, ipyw.HBox):
    pass

class VBox(WidgetContainer, ipyw.VBox):
    pass

# Cell
class WidgetList(WidgetContainer, ItemList, ipyw.Box):
    pass

class WidgetRow(WidgetList, ipyw.HBox):
    pass

class WidgetColumn(WidgetList, ipyw.VBox):
    pass

# Cell
class SelectionContainer(Selection, Widget):

    children = ipyw.trait_types.TypedTuple(help="List of widget children").tag(sync=True, **ipyw.widget.widget_serialization)

    def __init__(self, children=None, **kwargs):
        # because there is an optional name parameter from Selection parent and Widget
        if 'name' in kwargs:
            self.name = kwargs['name']
            del kwargs['name']
        if not children:
            children = tuple()
        super().__init__(children, **kwargs)

class SelectionBox(SelectionContainer, ipyw.Box):
    pass

class SelectionVBox(SelectionContainer, ipyw.VBox):
    pass

class SelectionHBox(SelectionContainer, ipyw.HBox):
    pass

# Cell
class TabAcc(Collection, Widget, ipyw.widget_selectioncontainer._SelectionContainer):

    children = ipyw.trait_types.TypedTuple(trait=Instance(Widget), help="List of widget children").tag(sync=True, **ipyw.widget.widget_serialization)
    selected_name = Unicode(default_value=None, allow_none=True)
    selected = Any(default_value=None, allow_none=True)

    def __init__(self, children=None, **kwargs):
        super().__init__(**kwargs)
        self._names = {}
        self._indices = {}
        if children:
            self.children = children

    def _add_child(self, child):
        assert hasattr(child, 'name'), 'Children of Tab and Accordion must have a name attribute'
        self._names[self._size] = child.name
        self._indices[child.name] = self._size
        super()._add_child(child)

        self.set_title(self._size - 1, child.name)

    @observe('selected_index')
    def _observe_selected_index(self, change):
        if change['new']  is not None:
            self.selected_name = self._names[change['new']]
        else:
            self.selected_name = None
        return change['new']

    @observe('selected_name')
    def _observe_selected_name(self, change):
        if change['new'] is not None:
            self.selected_index = self._indices[change['new']]
        else:
            self.selected_index = None
        return change['new']

    @observe('_children')
    def _observe_children(self, change):
        self.children = change['new']
        return change['new']

class Tab(TabAcc, ipyw.Tab):
    pass

class Accordion(TabAcc, ipyw.Accordion):
    pass


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/views.ratio.ipynb (unless otherwise specified).

__all__ = ['RatioLandingView', 'RatioLanding', 'ColorInt', 'RatioEstimatesView', 'RatioEstimates']

# Cell
from ..items import ItemList, ItemArray
from ..widget_containers import VBox, HBox
from ..widgets import Text, Dropdown, Button, Grid, FloatText, AlignedLabel, IntLabel, IntText
from ..rpy import DTYPES
from ..models.ratio import RatioEstimatesModel, Ratios
from ..mvc import Controller

# Cell
import traitlets
from traitlets import Unicode, observe, validate, Int
import ipywidgets as ipyw
import numpy as np

# Cell
class RatioLandingView(VBox):

    def __init__(self, rowMax=6, colMax=6, **kwargs):
        super().__init__(**kwargs)

        ''' rowBox widgets'''
        self.rowName = Text(description='Factor:', layout=ipyw.Layout( max_width = '160px'))
        options=[i for i in range(2, rowMax + 1)]
        self.n_row = Dropdown(description='Levels:', layout=ipyw.Layout(width='135px'), options=options)

        rowLevels = [Text(layout=ipyw.Layout(max_width = '62px')) for i in range(rowMax)]
        rowLevels[0].description = 'Names:'
        rowLevels[0].layout.max_width = '150px'
        self.rowLevels = ItemList(rowLevels)

        ''' columnBox widgets '''
        self.colName = Text(description='Factor:', layout=ipyw.Layout( max_width = '160px'))
        options = [i for i in range(2, colMax + 1)]
        self.n_col  = Dropdown(description='Levels:', layout=ipyw.Layout(width='135px'), options=options)

        colLevels = [Text(layout=ipyw.Layout(max_width = '62px')) for i in range(colMax)]
        colLevels[0].description = 'Names:'
        colLevels[0].layout.max_width = '150px'
        self.colLevels = ItemList(colLevels)

        ''' add children '''
        self.rowBox = HBox()
        self.rowBox.children += (self.rowName, self.n_row) + self.rowLevels.children
        self.colBox = HBox()
        self.colBox.children += (self.colName, self.n_col) + self.colLevels.children
        self.children = (self.rowBox, self.colBox)

# Cell
class RatioLanding( Controller, RatioLandingView):

    def __init__(self, model, **kwargs):
        super().__init__(model.rowMax, model.colMax,**kwargs)

        self.model = model
        self.initialize()

        ''' link widgets traits to model '''
        traitlets.link((self.model, 'rowName'), (self.rowName, 'value'))
        traitlets.link((self.model, 'colName'), (self.colName, 'value'))
        traitlets.link((self.model, 'n_row'), (self.n_row, 'value'))
        traitlets.link((self.model, 'n_col'), (self.n_col, 'value'))

        ''' observe'''
        self.n_row.observe(self.observe_n_row, names='value')
        self.n_col.observe(self.observe_n_col, names='value')
        self.rowLevels.observe(self.changeRowLevels, 'value', type='child_change')


    def initialize(self):
        self.rowName.value = self.model.rowName
        self.colName.value = self.model.colName
        self.n_row.value = self.model.n_row
        self.n_col.value = self.model.n_col
        self.rowLevels.set_list(self.model.rowLevels)
        self.colLevels.set_list(self.model.colLevels)

    def observe_n_row(self, change):
        for i in range(2, change['new']):
            self.rowLevels[i].visible()
        for i in range(change['new'], 6):
            self.rowLevels[i].invisible()
        return change

    def observe_n_col(self, change):
        for i in range(2, change['new']):
            self.colLevels[i].visible()
        for i in range(change['new'], 6):
            self.colLevels[i].invisible()
        return change

    def changeRowLevels(self, change):
        i = change['index']
        value = change['new']
        self.model.rowLevels[i] = value
        self.rowLevels[i].value = value

# Cell
class ColorInt(Button):

    value = Int(0)
    color = Unicode('none')

    def __init__(self, width='auto', height='auto', **kwargs):
        super().__init__(**kwargs)
        self.description = str(self.value)
        self.on_click(self._on_click)

    @observe('value')
    def _on_value_change(self, proposal):
        self.description = str(self.value)

    @validate('color')
    def _valid_value(self, proposal):
        if proposal['value'] not in ('red', 'green', 'none'):
            raise TraitError('Invalid color: Valid colors are red, green, and none')
        return proposal['value']

    @observe('color')
    def _observe_color(self, b):
        if self.color == 'red':
            self.layout.border='3px solid red'
        elif self.color == 'green':
            self.layout.border='3px solid green'
        elif self.color == 'none':
            self.layout.border='none'

    def _on_click(self, button):
        if button.color == 'red':
            button.color = 'green'
        elif button.color == 'green':
            button.color = 'none'
        elif button.color == 'none':
            button.color = 'red'

# Cell
class RatioEstimatesView(Grid):

    def __init__(self, n_row=6, n_col=6, decimals=3, **kwargs):
        grid_rows = int(n_row+3)
        grid_cols = int(n_col+3)
        super().__init__(grid_rows, grid_cols, width='100%', **kwargs)


        '''layouts'''
        # note: min_width prevents overflow
        # https://css-tricks.com/equal-width-columns-in-css-grid-are-kinda-weird/
        layout = {'min_width':'60px', 'width':'auto', 'height':'auto'}

        rowRatios = [FloatText(#step=1/10**self.model.rowRatios.decimals, value=self.model.rowRatios[i],
                               layout=ipyw.Layout(**layout)) for i in range(n_row)]
        self.rowRatios = ItemList(rowRatios)
        self.add_widget_list(self.rowRatios, 2, 0, direction='col')

        colRatios = [FloatText(#step=1/10**self.model.colRatios.decimals, value=self.model.colRatios[i],
                               layout=ipyw.Layout(**layout)) for i in range(n_col)]
        self.colRatios = ItemList(colRatios)
        self.add_widget_list(self.colRatios, 0, 2, direction='row')


        colorIntArray = [ColorInt(layout=ipyw.Layout(**layout)) for j in range(n_col) for i in range(n_row)]
        self.colorIntArray = ItemArray(n_row, n_col, colorIntArray, traits=('value', 'color'))
        self.add_widget_array(self.colorIntArray, 2, 2)

        widget_list = [AlignedLabel(layout=ipyw.Layout(**layout)) for i in range(n_row)]
        self.rowLevelNames = ItemList(widget_list)
        self.add_widget_list(self.rowLevelNames, 2, 1, direction='col')

        widget_list = [AlignedLabel(layout=ipyw.Layout(**layout)) for i in range(n_col)]
        self.colLevelNames = ItemList(widget_list)
        self.add_widget_list(self.colLevelNames, 1, 2, direction='row')

        widget_list = [IntLabel(layout=ipyw.Layout(**layout)) for i in range(n_row)]
        self.rowTotals = ItemList(widget_list)
        self.add_widget_list(self.rowTotals, 2, n_col + 2, direction='col')


        widget_list = [IntLabel(layout=ipyw.Layout(**layout)) for i in range(n_col)]
        self.colTotals = ItemList(widget_list)
        self.add_widget_list(self.colTotals, n_row + 2, 2, direction='row')

        self.total = IntText(layout=ipyw.Layout(**layout))
        self.add_widget_list([self.total], n_row + 2, n_col+2)

# Cell
class RatioEstimates(RatioEstimatesView, Controller):

    def __init__(self, model, **kwargs):
        super().__init__(model.n_row, model.n_col, **kwargs)

        self.model = model
        self.initialize()

        ''' link '''
        traitlets.link((self.model, 'total'), (self.total, 'value'))

        ''' observe '''
        self.total.observe(self.observeTotal)
        self.rowRatios.observe(self.observeRowRatios, type='child_change')
        self.colRatios.observe(self.observeColRatios, type='child_change')

    def initialize(self):
        for ratio in self.rowRatios:
            ratio.step = self.model.rowRatios.step
        for ratio in self.colRatios:
            ratio.step = self.model.colRatios.step
        self.rowRatios.unobserve(self.observeRowRatios, type='child_change')
        self.rowRatios.set_list(self.model.rowRatios)
        self.rowRatios.observe(self.observeRowRatios, type='child_change')
        self.colRatios.unobserve(self.observeColRatios, type='child_change')
        self.colRatios.set_list(self.model.colRatios)
        self.colRatios.observe(self.observeColRatios, type='child_change')
        self.colorIntArray.set_array(self.model.ns_cells, 'value')
        self.colorIntArray.set_array(self.model.colors, 'color')
        self.rowLevelNames.set_list(self.model.rowLevelNames)
        self.colLevelNames.set_list(self.model.colLevelNames)
        self.rowTotals.set_list(self.model.rowTotals)
        self.colTotals.set_list(self.model.colTotals)

    def onAnyChange(self):
        self.colorIntArray.set_array(self.model.ns_cells)
        self.rowTotals.set_list(self.model.rowTotals)
        self.colTotals.set_list(self.model.colTotals)

    def observeTotal(self, change):
        self.onAnyChange()

    def observeRowRatios(self, change):
        self.model.incRowRatio(change['index'], change['new'] - change['old'])
        self.rowRatios.unobserve_children('value')
        self.rowRatios.set_list(self.model.rowRatios)
        self.rowRatios.observe_children('value')
        self.onAnyChange()

    def observeColRatios(self, change):
        self.model.incColRatio(change['index'], change['new'] - change['old'])
        self.colRatios.unobserve_children('value')
        self.colRatios.set_list(self.model.colRatios)
        self.colRatios.observe_children('value')
        self.onAnyChange()


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/06_traits.ipynb (unless otherwise specified).

__all__ = ['IntArray', 'FloatArray', 'Int', 'Float', 'LinkAllMixin']

# Cell
from traittypes import Array
from traitlets import TraitType, HasTraits, Unicode
import traitlets
import numpy as np

# Cell
from .rpy import DTYPES, AS_DTYPE

# Cell
class IntArray(Array):
    def __init__(self, default_value=None, size=None, **kwargs):
        if size:
            assert 'default_value' not in kwargs, "Only one of 'size' or 'default_value' should be provided"
            kwargs['default_value'] = np.empty(size, dtype=DTYPES['int'])
        if default_value is not None:
            kwargs['default_value'] = default_value
        super().__init__(dtype=DTYPES['int'], **kwargs)

# Cell
class FloatArray(Array):
    def __init__(self, default_value=None, size=None, **kwargs):
        if size:
            assert 'default_value' not in kwargs, "Only one of 'size' or 'default_value' should be provided"
            kwargs['default_value'] = np.empty(size, dtype=DTYPES['float'])
        if default_value is not None:
            kwargs['default_value'] = default_value
        super().__init__(dtype=DTYPES['float'], **kwargs)

# Cell
class Int(TraitType):

    info_text = 'an integer compatible with R and numpy'

    def validate(self, obj, value):
        if isinstance(value, int) or (hasattr(value, 'dtype') and np.issubdtype(value.dtype, np.integer)):
            return AS_DTYPE['int'](value)
        self.error(obj, value)

# Cell
class Float(TraitType):

    info_text = 'an integer compatible with R and numpy'

    def validate(self, obj, value):
        if isinstance(value, int) or isinstance(value, float) or (hasattr(value, 'dtype') and np.issubdtype(value.dtype, np.floating)):
            return AS_DTYPE['float'](value)
        self.error(obj, value)

# Cell
class LinkAllMixin():

    def link_all(self, source, destination):
        dest_traits = destination.traits()
        for trait in source.traits():
            if trait in dest_traits:
                traitlets.link((source, trait), (destination, trait))


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/models.chisq.ipynb (unless otherwise specified).

__all__ = ['ChiSqLandingModel', 'ChiSqModel']

# Cell
from ..superpower.chisq import ChiSqCon, ChiSqInd
from .corr import CramerVModel, CramerVLandingModel
from ..rpy import RMixins

# Cell
class ChiSqLandingModel(CramerVLandingModel):
    pass

# Cell
class ChiSqModel(CramerVModel, RMixins):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.ind = ChiSqInd()
        self.con = ChiSqCon()

    def run(self):
        if self.hasColors:
            res = self.runEffectSize()
            w = res['df'].power[0]
            kwargs = self.getTraits(['mata', 'matb', 'rcomp', 'pVal'])
            nrow = self.matrix.shape[0]
            kwargs['mata'] = self.matrix_func(kwargs['mata'], nrow=nrow)
            kwargs['matb'] = self.matrix_func(kwargs['matb'], nrow=nrow)
            return self.con.run(**kwargs)
        else:
            kwargs = self.getTraits(['ns_cells', 'pVal', 'n_col', 'n_row', 'w'])
            kwargs['n_row'], kwargs['n_col'] = kwargs['ns_cells'].shape
            kwargs['ns_cells'] = kwargs['ns_cells'].flatten()
            return self.ind.run(**kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/views.table_widget.ipynb (unless otherwise specified).

__all__ = ['logger', 'Table']

# Cell
from traitlets import Unicode, Bool, validate, TraitError, Instance, Dict, List, Unicode, Int
from ipywidgets import DOMWidget, register
import ipywidgets as ipyw

# Cell
from ..config import APP_DIR, CONFIG
from ..log import HANDLERS
from ..widgets import Widget
logger = HANDLERS.getLogger(__name__)

# Cell
@register
class Table(DOMWidget, Widget):
    _view_name = Unicode('TableView').tag(sync=True)
    _view_module = Unicode('widgets/table_widget').tag(sync=True)
    _view_module_version = Unicode('0.1.0').tag(sync=True)

    # Attributes
    data = List([]).tag(sync=True)
    selected = Int(0).tag(sync=True)

    @validate('data')
    def _valid_value(self, proposal):
        for row in proposal['value']:
            if type(row) is not dict:
                raise TraitError('All list items should be dictionaries')
            required_keys = ['id', 'name', 'runtime', 'runNum']
            for key in required_keys:
                if key not in row:
                    raise TraitError('All table rows must contain the key \"' + key + '\"')
        return proposal['value']


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/superpower.chisq.ipynb (unless otherwise specified).

__all__ = ['ChiSqInd', 'ChiSqCon']

# Cell
from ..rpy import RFunction
import numpy as np

# Cell
class ChiSqInd(RFunction):

    argNames = ['n_row', 'n_col', 'N', 'ns_cells', 'pVal', 'w']

    def __init__(self, argNames=argNames, **kwargs):
        super().__init__(name='powerChiSqInd', argNames=argNames, **kwargs)

class ChiSqCon(RFunction):

    argNames=['pVal', 'mata', 'matb', 'rcomp']

    def __init__(self, argNames=argNames, **kwargs):
        super().__init__(name='powerChiSqCont', argNames=argNames, **kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/settings.ipynb (unless otherwise specified).

__all__ = ['SettingsModel', 'SettingsView', 'Settings']

# Cell
from .mvc import Model, Controller
from .traits import Float
from .widget_containers import Box
from .widgets import Dropdown, FloatText

# Cell
from traitlets import Unicode, Float
import traitlets
import ipywidgets as ipyw

# Cell
class SettingsModel(Model):

    tail = Unicode()
    alpha = Float()

    def setLanding(self):
        self.tail = 'two'
        self.alpha = 0.05

# Cell
class SettingsView(Box):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.layout = ipyw.Layout(width='100%', flex_flow='wrap')

        tailOptions = [('2-tailed', 'two'), ('1-tailed, less', 'less'), ('1-tailed, greater', 'greater')]
        self.tail = Dropdown(options=tailOptions, description = 'Tail:')
        self.alpha = FloatText(description='α: ', step=0.01, min = 0.0, max = 1.0)
        self.children += (self.tail, self.alpha)

# Cell
class Settings(SettingsView):

    def __init__(self, model, **kwargs):
        super().__init__(**kwargs)
        self.model = model

        traitlets.link((self.model, 'tail'), (self.tail, 'value'))
        traitlets.link((self.model, 'alpha'), (self.alpha, 'value'))


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05_rpy.ipynb (unless otherwise specified).

__all__ = ['logger', 'RConsole', 'R_CONSOLE', 'rpy2', 'lib_loc', 'lib_paths', 'pkg_name', 'utils', 'types', 'DTYPES',
           'AS_DTYPE', 'Result', 'RFunction', 'RMixins']

# Cell
import time, os, datetime
import pandas as pd
import numpy as np
from IPython.display import display, Javascript, HTML
from collections import UserDict

# Cell
from .log import HANDLERS
from .config import CONFIG, APP_DIR
logger = HANDLERS.getLogger(__name__)

# Cell
class RConsole():

    def __init__(self):
        self.logger = HANDLERS.getLogger('r.console')
        self.setup()

    def print(self, x):
        self.logger.info(x)

    def warnError(self, x):
        self.logger.error(x)

    def setup(self):
        # output from the R console will now be handled by the above functions
        import rpy2
        from rpy2 import rinterface, robjects, situation
        rpy2.rinterface.initr()
        from rpy2.robjects import packages, numpy2ri, pandas2ri, conversion
        rpy2.robjects.numpy2ri.activate()
        rpy2.robjects.numpy2ri.activate()
        self.rpy2 = rpy2
        rpy2.rinterface_lib.callbacks.consolewrite_warnerror = self.warnError
        rpy2.rinterface_lib.callbacks.consolewrite_print = self.print
        self.cat = rpy2.rinterface.baseenv['cat']
        self.cat('RConsole initialized.')


R_CONSOLE = RConsole()
R_CONSOLE.setup()
rpy2 = R_CONSOLE.rpy2

# Cell
lib_loc = CONFIG['R']['lib_loc']
lib_paths = list(rpy2.robjects.r['.libPaths']())
if lib_loc:
    lib_loc = os.path.expanduser(lib_loc)
    lib_paths.insert(0, lib_loc)

# Cell
pkg_name='Superpower'
utils = rpy2.robjects.packages.importr('utils')
for lib_path in lib_paths:
    try:
        logger.info('Attempting Superpower import from %s...' % lib_path)
        SUPERPOWER = rpy2.robjects.packages.importr(pkg_name, lib_loc)
        if SUPERPOWER:
            logger.info('Import successful!')
            break
    except Exception as e:
        logger.exception('Import failed with message: %s' % str(e))
        logger.info('Attempting install from source to %s' % lib_loc)
        if CONFIG['R']['source']:
            source_path = os.path.expanduser(CONFIG['R']['source'])
            if not os.path.isabs(source_path):
                source_path = os.path.join(APP_DIR, source_path)
            if os.path.isfile(source_path):
                logger.info('Using source file %s' % source_path)
            else:
                raise Exception("Source file not found at %s" % source_path)
        else:
            raise Exception('Superpower not installed and source file not provided.')
        try:
            kwargs = {}
            kwargs['pkgs'] = source_path
            kwargs['repos'] = rinterface.NULL
            kwargs['type'] = 'source'
            kwargs['lib'] = lib_loc
            utils.install_packages(**kwargs)
            SUPERPOWER = rpy2.robjects.packages.importr(pkg_name, lib_loc)
        except Exception as e:
            logger.exception('Install failed with message: %s' % str(e))
            raise Exception('Fatal. Superpower could not be imported or installed.')
        if SUPERPOWER:
            logger.info('Install successful!')
        else:
            raise Exception('Install failed.')

# Cell
types = (('int', rpy2.rinterface.IntSexpVector([])), ('bool', rpy2.rinterface.BoolSexpVector([])), ('string', rpy2.rinterface.StrSexpVector([])), ('float', rpy2.rinterface.FloatSexpVector([])))
DTYPES = {}
AS_DTYPE = {}
for py_type, r_vector in types:
    np_array = np.asarray(r_vector)
    DTYPES[py_type] = np_array.dtype
    AS_DTYPE[py_type] = getattr(np, str(np_array.dtype))

# Cell
class Result(UserDict):

    def _repr_pretty_(self, p, cycle):
        with p.group(4, self.data['name'] +'([', '])'):
            p.breakable()
            for key, value in self.data.items():
                if key != 'name':
                    p.text(key  + ': ')
                    if isinstance(value, dict):
                        with p.group(4, '{', '}'):
                            for k, v in value.items():
                                p.text(k + ': ')
                                p.pretty(v)
                                p.text(', ')
                                p.breakable()
                    elif isinstance(value, pd.DataFrame):
                        p.pretty('See above data frame.')
                        display(HTML(value.to_html()))
                    else:
                        p.pretty(value)
                        p.text(',')
                    p.breakable()

# Cell
class RFunction():
    ''' The class used to create a model for any R function, and only one R function.
    This class abstracts away all rpy2 calls to make interaction with the SUPERPOWER library
    easy.
    '''

    def __init__(self, name, argNames=None):
        self.name = name
        self.function = getattr(SUPERPOWER, name)
        self.defaults = self.getDefaultKwargs()
        self.setArgNames(argNames)

    def setArgNames(self, argNames=None):
        '''Sets the names of all accepted arguments. Overrides previous.'''
        if argNames is None:
            self.getDefaultArgNames(self)
        else:
            if all(name in self.argNames for name in argNames):
                self.argNames = argNames
            else:
                raise Exception('All argNames must be valid arguments for the given R function')

    def setKwargs(self, **kwargs):
        """Adds kwargs if and only if the key is is a valid function argument.
        """
        self.kwargs = {k:v for (k, v) in kwargs.items() if k in self.argNames}

    def setDefaultArg(self, argName):
        self.kwargs[argName] = self.defaults[argName]

    def getDefaultArgNames(self):
        '''Gets the names of all accepted arguments'''
        formals = rpy2.robjects.r['formals']
        default = formals(self.function)
        return default.names

    def getDefaultKwargs(self):
        '''This method gets a dictionary of default arguments provided by the R function'''
        formals = rpy2.robjects.r['formals']
        default = formals(self.function)
        args = {}
        for i, name in enumerate(default.names):
            if len(default[i]) > 0 and default[i][0] != rpy2.robjects.r("NULL"):
                ary = np.asarray(default[i])
                if len(ary) > 1:
                    args[name] = ary
                else:
                    args[name] = ary[0]
        return args

    def run(self, alert=True, **kwargs):
        '''This method runs the R function associated with the model. The run will be
        timed and the runtime and results are added to the results dictionary. You should
        not have to modify this function to run your custom Model
        '''
        if kwargs:
            self.setKwargs(**kwargs)
        try:
            logger.info('Running %s with kwargs...' % self.name)
            for key, value in self.kwargs.items():
                logger.info('\t%s: %s' % (key, str(value).replace("\n", "")))
            ret = {}
            startTime = time.time()
            ret['name'] = self.name
            ret['kwargs'] = self.kwargs
            ret['df'] = self.function(**self.kwargs)
            if isinstance(ret['df'], rpy2.robjects.vectors.ListVector):
                ret['df'] = self.listVector_to_DataFrame(ret['df'])
            ret['df'] = self.convert_to_pandas(ret['df'])
            ret['runtime'] = (time.time() - startTime)
            ret['id'] = int(round(time.time() * 1000))
            ret['date'] = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            logger.info('ID: %i, Date: %s, Runtime: %s' % (ret['id'], str(ret['date']), str(ret['runtime'])))
            logger.info('Results:\n%s' % ret['df'].to_string())
            return Result(ret)
        except Exception as e:
            if alert:
                e = str(e).replace('\n', ' ').replace('\n', '').replace('"','\\"')
                display(Javascript('alert("' + e + '");'))
                return None
            else:
                raise e

    def listVector_to_DataFrame(self, result):
        df_dict = {}
        for i in range(len(result.names)):
            df_dict[result.names[i]] = result[i]
        return rpy2.robjects.DataFrame(df_dict)

    def convert_to_pandas(self, df):
        with rpy2.robjects.conversion.localconverter(rpy2.robjects.default_converter + rpy2.robjects.pandas2ri.converter):
            return rpy2.robjects.conversion.rpy2py(df)

    def attr_dict(self):
        keys = ['name', 'result', 'packageName', 'runtime', 'id', 'kwargs', 'defaults']
        return {key: value for key, value in self.__dict__.items() if key in keys}

    def _repr_pretty_(self, p, cycle):
        with p.group(4, str(self.name) +'([', '])'):
            p.breakable()
            for key, value in self.attr_dict().items():
                if key != 'name':
                    p.text(key  + ': ')
                    if isinstance(value, dict):
                        with p.group(4, '{', '}'):
                            for k, v in value.items():
                                p.text(k + ': ')
                                p.pretty(v)
                                p.text(',')
                                p.breakable()
                    else:
                        p.pretty(value)
                        p.text(',')
                        p.breakable()
                    p.breakable()

# Cell
class RMixins():

    def __init__(self):
        self.matrix_func = rpy2.robjects.baseenv['matrix']


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/07_mvc.ipynb (unless otherwise specified).

__all__ = ['Model', 'Controller', 'MiniView']

# Cell
from traitlets import HasTraits, Int, Unicode, List, observe, link
import ipywidgets as ipyw

# Cell
from .widget_containers import HBox, WidgetList
from .widgets import Text, Dropdown

# Cell
class Model(HasTraits):

    def __init__(self, **kwargs):
        if not kwargs:
            self.setLanding()
        else:
            self.setTraits(**kwargs)
        super().__init__(**kwargs)

    def initTraits(self):
        for trait in self.traits().keys():
            getattr(self, trait)

    def setTraits(self, **kwargs):
        with self.hold_trait_notifications():
            for key, value in kwargs.items():
                setattr(self, key, value)

    def getTraits(self, traits=None):
        ret = {}
        if not traits:
            traits = self.traits().keys()
            self.initTraits()
        for trait in traits:
            ret[trait] = self._trait_values[trait]
        return ret

    def run(self):
        raise NotImplementedError('All Model subclasses must implement a run method')

    def setLanding(self, **kwargs):
        raise NotImplementedError('All Model subclasses must implement a setLanding method')

    def setEstimates(self, **kwargs):
        raise NotImplementedError('All Model subclasses must implement a setEstimates method')

    def _repr_pretty_(self, p, cycle):
        class_name = self.__class__.__name__
        with p.group(4, str(class_name) +'(', ')'):
            p.breakable()
            for trait, value in self._trait_values.items():
                if not trait.startswith('_'):
                    p.text(trait + ': ')
                    p.pretty(value)
                    p.text(',')
                    p.breakable()

# Cell
class Controller():

    def initialize(self):
        raise NotImplementedError()

    def setLanding(self):
        self.model.setLanding()
        self.initialize()

    def setEstimates(self):
        self.model.setEstimates()
        self.initialize()

# Cell
class MiniView(HBox):

    def __init__(self, model, **kwargs):
        super().__init__(**kwargs)

        self.model = model

        ''' layouts '''
        nameLayout = ipyw.Layout( max_width = '160px')
        levelsLayouts = [ipyw.Layout( max_width = '62px') for i in range(0, self.model.max_n)]
        levelsLayouts[0].max_width  = '150px'
        dropdownLayout = ipyw.Layout(width='135px')

        ''' rowBox widgets'''
        self.name = Text(description='Name:', layout=nameLayout, value = self.model.name)
        options=[i for i in range(2, self.model.max_n + 1)]
        self.n = Dropdown(description='Number:', layout=ipyw.Layout(width='135px'), options=options, value=self.model.n)

        levels = [ Text(description = 'Levels:', layout=levelsLayouts[0], value=self.model.levels[0])]
        levels += [ Text(layout=levelsLayouts[i], value=self.model.levels[i]) for i in range(1, self.model.n)]
        levels += [ Text(layout=levelsLayouts[i]) for i in range(self.model.n, self.model.max_n)]
        self.levels = WidgetList(levels)

        levelNames = [ Text(description = 'LevelNames:', layout=levelsLayouts[0], value=self.model.levels[0])]
        levelNames += [ Text(layout=levelsLayouts[i], value=self.model.levels[i]) for i in range(1, self.model.n)]
        levelNames += [ Text(layout=levelsLayouts[i]) for i in range(self.model.n, self.model.max_n)]
        self.levelNames = WidgetList(levelNames)

        self.children = (self.name, self.n) + self.levels.children + self.levelNames.children


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/views.chisq.ipynb (unless otherwise specified).

__all__ = ['ChiSqLanding', 'ChiSq']

# Cell
from .ratio import RatioLanding, RatioEstimates
from ..widgets import Textarea, Button
from ..widget_containers import  HBox, VBox
from ..models.chisq import ChiSqModel
from ..mvc import Controller
from .corr import CramerV

# Cell
import ipywidgets as ipyw

# Cell
class ChiSqLanding(RatioLanding):
    pass

# Cell
class ChiSq(CramerV):

    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04_log.ipynb (unless otherwise specified).

__all__ = ['DispatchingFormatter', 'WidgetHandler', 'Handlers', 'HANDLERS', 'logger']

# Cell
import logging
import ipywidgets as ipyw
import traitlets
from IPython.display import display
import warnings
import os, pathlib, urllib
from IPython.display import HTML, Javascript
from pathlib import Path

# Cell
from .config import CONFIG, APP_DIR
from .widgets import Textarea

# Cell
class DispatchingFormatter:
    """Dispatch formatter for logger and it's sub logger."""
    def __init__(self):
        self.formatters = {
            'r.console': logging.Formatter('[R.%(levelname)s] %(message)s'),
            #'js.output': logging.Formatter('[JS.OUTPUT] %(message)s'),
            #'js': logging.Formatter('[JS.%(levelname)s] %(message)s'),
            'py.warnings': logging.Formatter('[PY.WARNINGS] %(message)s'),
            '__main__': logging.Formatter('[PY.%(levelname)s] %(message)s'),
            'default': logging.Formatter('[PY.%(levelname)s](%(name)s:%(lineno)d)  %(message)s')
        }

    def format(self, record):
        formatter = self.formatters.get(record.name, self.formatters['default'])
        return formatter.format(record)

# Cell
class WidgetHandler(Textarea, logging.Handler):

    def __init__(self, **kwargs):
        super().__init__(layout={'width': '100%', 'height': '250px'}, disabled=True, **kwargs)

    def emit(self, record):
        self.value += str(self.format(record)) + '\n'

# Cell
class Handlers():

    def __init__(self, level, filename, mode):
        self.formatter = DispatchingFormatter()
        self.handlers = {}
        self.level = getattr(logging, level)
        self.logger = self.getLogger(__name__)

        ''' capture warnings '''
        logging.captureWarnings(True)
        self.py_warnings_logger = logging.getLogger('py.warnings')

        ''' add file handler '''
        self.addHandler('file', logging.FileHandler(filename, mode))
        self.logger.info('logging.captureWarnings(True)')


        ''' need to be loaded '''
        self.loaded = False
        self.requireBaseUrl = None

    def addHandler(self, name, handler):
        handler.setFormatter(self.formatter)
        self.py_warnings_logger.addHandler(handler)
        self.handlers[name] = handler
        self.logger = self.getLogger(__name__)

    def addWidget(self, widget):
        self.addHandler('widget', widget)

    def loadCSS(self):
        css_path = os.path.join(APP_DIR, 'assets/css')
        self.logger.info('Loading css files...')
        for file in list(Path(css_path).rglob('*.css')):
            with open(file, 'r') as f:
                try:
                    filestr = '<style>\n' + f.read() + '\n</style>'
                    display(HTML(filestr))
                    self.logger.info('\t%s' % file)
                except:
                    self.logger.exception('Error loading css file %s' % file)

    def _ipython_display_(self):
        if 'widget' in self.handlers:
            display(self.handlers['widget'])
            if not self.loaded:
                self.loadCSS()

    def getLogger(self, name='__main__'):
        logger = logging.getLogger(name)
        logger.setLevel(self.level)
        for handler in self.handlers.values():
            logger.addHandler(handler)
        return logger

# Cell
HANDLERS = Handlers(CONFIG['LOG']['level'], CONFIG['LOG']['filename'], CONFIG['LOG']['mode'])
HANDLERS.addWidget(WidgetHandler())
logger = HANDLERS.getLogger(__name__)
logger.info('Welcome to Superpower!')


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/views.tabs.ipynb (unless otherwise specified).

__all__ = ['LandingTab', 'EstimatesTab', 'ResultsTab', 'SettingsTab']

# Cell
from ..widget_containers import VBox, HBox

# Cell
class LandingTab(VBox):
    def __init__(self, dropdown, landing_box, okay_button, **kwargs):
        super().__init__(**kwargs)
        self.children = (dropdown, landing_box, okay_button)

# Cell
class EstimatesTab(VBox):
    def __init__(self, estimates_box, run_button, **kwargs):
        super().__init__(**kwargs)
        self.children = (estimates_box, run_button)

# Cell
class ResultsTab(VBox):
    def __init__(self, reset_button, repopulate_landing, repopulate_estimates, resultsView, **kwargs):
        super().__init__(**kwargs)
        hbox = HBox()
        hbox.children = (reset_button, repopulate_landing, repopulate_estimates)
        self.children = (resultsView, hbox)

# Cell
class SettingsTab(VBox):
    def __init__(self, settingsView, log_widget, **kwargs):
        super().__init__(**kwargs)
        self.children = (settingsView, log_widget)


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"Item": "00_items.ipynb",
         "Collection": "00_items.ipynb",
         "Selection": "controller.ipynb",
         "ItemList": "00_items.ipynb",
         "ItemArray": "00_items.ipynb",
         "Widget": "01_widgets.ipynb",
         "Output": "01_widgets.ipynb",
         "IntSlider": "01_widgets.ipynb",
         "FloatSlider": "01_widgets.ipynb",
         "IntProgress": "01_widgets.ipynb",
         "BoundedIntText": "01_widgets.ipynb",
         "BoundedFloatText": "01_widgets.ipynb",
         "IntText": "01_widgets.ipynb",
         "FloatText": "01_widgets.ipynb",
         "ToggleButtons": "01_widgets.ipynb",
         "Checkbox": "01_widgets.ipynb",
         "Dropdown": "01_widgets.ipynb",
         "RadioButtons": "01_widgets.ipynb",
         "Select": "01_widgets.ipynb",
         "Text": "01_widgets.ipynb",
         "Textarea": "01_widgets.ipynb",
         "Label": "01_widgets.ipynb",
         "HTML": "01_widgets.ipynb",
         "HTMLMath": "01_widgets.ipynb",
         "Image": "01_widgets.ipynb",
         "Button": "01_widgets.ipynb",
         "ColorPicker": "01_widgets.ipynb",
         "Layout": "01_widgets.ipynb",
         "HBox": "02_widget_containers.ipynb",
         "AlignedLabel": "01_widgets.ipynb",
         "IntLabel": "01_widgets.ipynb",
         "Grid": "01_widgets.ipynb",
         "WidgetContainer": "02_widget_containers.ipynb",
         "Box": "02_widget_containers.ipynb",
         "VBox": "02_widget_containers.ipynb",
         "WidgetList": "02_widget_containers.ipynb",
         "WidgetRow": "02_widget_containers.ipynb",
         "WidgetColumn": "02_widget_containers.ipynb",
         "SelectionContainer": "02_widget_containers.ipynb",
         "SelectionBox": "02_widget_containers.ipynb",
         "SelectionVBox": "02_widget_containers.ipynb",
         "SelectionHBox": "02_widget_containers.ipynb",
         "TabAcc": "02_widget_containers.ipynb",
         "Tab": "02_widget_containers.ipynb",
         "Accordion": "02_widget_containers.ipynb",
         "APP_DIR": "03_config.ipynb",
         "CONFIG": "03_config.ipynb",
         "USER_DIR": "03_config.ipynb",
         "DispatchingFormatter": "04_log.ipynb",
         "WidgetHandler": "04_log.ipynb",
         "Handlers": "04_log.ipynb",
         "HANDLERS": "04_log.ipynb",
         "logger": "views.table_widget.ipynb",
         "RConsole": "05_rpy.ipynb",
         "R_CONSOLE": "05_rpy.ipynb",
         "rpy2": "results.ipynb",
         "lib_loc": "05_rpy.ipynb",
         "lib_paths": "05_rpy.ipynb",
         "pkg_name": "05_rpy.ipynb",
         "utils": "05_rpy.ipynb",
         "types": "05_rpy.ipynb",
         "DTYPES": "05_rpy.ipynb",
         "AS_DTYPE": "05_rpy.ipynb",
         "Result": "05_rpy.ipynb",
         "RFunction": "05_rpy.ipynb",
         "RMixins": "05_rpy.ipynb",
         "IntArray": "06_traits.ipynb",
         "FloatArray": "06_traits.ipynb",
         "Int": "06_traits.ipynb",
         "Float": "06_traits.ipynb",
         "LinkAllMixin": "06_traits.ipynb",
         "Model": "07_mvc.ipynb",
         "Controller": "controller.ipynb",
         "MiniView": "07_mvc.ipynb",
         "ChiSqLandingModel": "models.chisq.ipynb",
         "ChiSqModel": "models.chisq.ipynb",
         "CramerVLandingModel": "models.corr.ipynb",
         "AddStringsMixin": "models.corr.ipynb",
         "CramerVModel": "models.corr.ipynb",
         "TetraLandingModel": "models.corr.ipynb",
         "TetraModel": "models.corr.ipynb",
         "CorrLandingModel": "models.corr.ipynb",
         "CorrModel": "models.corr.ipynb",
         "SpearmansRhoModel": "models.corr.ipynb",
         "PearsonModel": "models.corr.ipynb",
         "PBiModel": "models.corr.ipynb",
         "Ratios": "models.ratio.ipynb",
         "RatioArray": "models.ratio.ipynb",
         "RatioLandingModel": "models.ratio.ipynb",
         "RatioEstimatesModel": "models.ratio.ipynb",
         "ResultsModel": "results.ipynb",
         "ResultsView": "results.ipynb",
         "Results": "results.ipynb",
         "SettingsModel": "settings.ipynb",
         "SettingsView": "settings.ipynb",
         "Settings": "settings.ipynb",
         "ChiSqInd": "superpower.chisq.ipynb",
         "ChiSqCon": "superpower.chisq.ipynb",
         "CorrPearson": "superpower.corr.ipynb",
         "CorrPBi": "superpower.corr.ipynb",
         "CorrSpearmansRho": "superpower.corr.ipynb",
         "CorrCramerV": "superpower.corr.ipynb",
         "CorrTetra": "superpower.corr.ipynb",
         "ChiSqLanding": "views.chisq.ipynb",
         "ChiSq": "views.chisq.ipynb",
         "CramerVLanding": "views.corr.ipynb",
         "CramerVEstimates": "views.corr.ipynb",
         "CramerV": "views.corr.ipynb",
         "TetraLanding": "views.corr.ipynb",
         "Tetra": "views.corr.ipynb",
         "CorrLanding": "views.corr.ipynb",
         "SpearmansRhoLanding": "views.corr.ipynb",
         "PBiLanding": "views.corr.ipynb",
         "PearsonLanding": "views.corr.ipynb",
         "CorrView": "views.corr.ipynb",
         "Corr": "views.corr.ipynb",
         "SpearmansRho": "views.corr.ipynb",
         "PBi": "views.corr.ipynb",
         "Pearson": "views.corr.ipynb",
         "RatioLandingView": "views.ratio.ipynb",
         "RatioLanding": "views.ratio.ipynb",
         "ColorInt": "views.ratio.ipynb",
         "RatioEstimatesView": "views.ratio.ipynb",
         "RatioEstimates": "views.ratio.ipynb",
         "Table": "views.table_widget.ipynb",
         "LandingTab": "views.tabs.ipynb",
         "EstimatesTab": "views.tabs.ipynb",
         "ResultsTab": "views.tabs.ipynb",
         "SettingsTab": "views.tabs.ipynb"}

modules = ["items.py",
           "widgets.py",
           "widget_containers.py",
           "config.py",
           "log.py",
           "rpy.py",
           "traits.py",
           "mvc.py",
           "controller.py",
           "models/chisq.py",
           "models/corr.py",
           "models/ratio.py",
           "results.py",
           "settings.py",
           "superpower/chisq.py",
           "superpower/corr.py",
           "views/chisq.py",
           "views/corr.py",
           "views/ratio.py",
           "views/table_widget.py",
           "views/tabs.py"]

doc_url = "https://pages.github.itap.purdue.edu/brewer36/superpower_gui/"

git_url = "https://github.itap.purdue.edu/superpower_gui/superpower_gui/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/models.ratio.ipynb (unless otherwise specified).

__all__ = ['Ratios', 'RatioArray', 'RatioLandingModel', 'RatioEstimatesModel']

# Cell
import numpy as np
from traitlets import List, Unicode, HasTraits, TraitType, observe, Int, Float
from traittypes import Array

# Cell
from ..rpy import DTYPES
from ..mvc import Model
from ..superpower.chisq import ChiSqInd

# Cell
class Ratios(np.ndarray):

    '''    Note that this doesn't support view casting and new-from-template. It just
    allows us to have access
    '''
    def __new__(subtype, array=None, size=3, decimals=3, **kwargs):
        if array is not None:
            array = np.array(array, dtype=np.dtype(float))
            size = array.size
        else:
            size = size
        precision = 10**decimals
        fill_value = round(int(1/size * precision)/precision, decimals)
        total = int(fill_value * precision) * size
        if array is not None:
            arrayTotal = int(np.sum(array) * precision)
            if not total == arrayTotal:
                raise Exception('Array does not contain valid sum %.3f' % round(total/precision, decimals))
            obj = array.view(subtype)
        else:
            obj = np.full(size, fill_value, dtype=np.dtype(float)).view(subtype)
        obj.precision = precision
        obj.decimals = decimals
        obj.total = total
        obj.step = 1/obj.precision
        obj.n = size
        obj.j = 0
        obj.k = 0
        return obj

    def __array_finalize__(self, obj):
        if obj is None: return
        self.precision = getattr(obj, 'precision', None)
        self.decimals = getattr(obj, 'decimals', None)
        self.total = getattr(obj, 'total', None)
        self.step = getattr(obj, 'step', None)
        self.n = getattr(obj, 'n', None)
        self.j = getattr(obj, 'j', None)
        self.k = getattr(obj, 'k', None)

    def __setitem__(self, i, value):
        diff = int(round(value - self[i], 3) * self.precision)
        if diff > 0:
            self.inc(i, diff)
        elif diff < 0:
            self.dec(i, abs(diff))

    def __add__(self, b):
        print('add: %s'% str(b))

    def __iadd__(self, b):
        print('addi: %s'% str(b))

    def __sadd__(self, b):
        print('adds: %s'% str(b))

    def inc(self, i, diff):
        change = 0
        nums = np.copy(self) * self.precision
        nums = nums.astype(int)
        if diff + nums[i] > self.total:
            diff = self.total - nums[i]
        while diff > 0:
            if self.j != i:
                if nums[self.j] - 1 >= 0:
                    nums[self.j] = nums[self.j] - 1
                    nums[i] = nums[i] + 1
                    diff -= 1
                    change += 1
            self._inc_j()
        for i, val in enumerate(np.around(nums/self.precision, 3)):
            self.itemset(i, val)

    def dec(self, i, diff):
        nums = np.copy(self) * self.precision
        nums = nums.astype(int)
        if nums[i] - diff < 0:
            diff = nums[i]
        while diff > 0:
            if self.k != i:
                if nums[self.k] + 1 <= self.total:
                    nums[self.k] = nums[self.k] + 1
                    nums[i] = nums[i] - 1
                    diff -= 1
            self._inc_k()
        for i, val in enumerate(np.around(nums/self.precision, 3)):
            self.itemset(i, val)

    def _inc_j(self):
        self.j = (self.j + 1) % self.n

    def _inc_k(self):
        self.k = (self.k + 1) % self.n

# Cell
class RatioArray(Array):

    info_text = 'a Ratio object'

    def __init__(self, default_value=None, size=3, decimals=3, **kwargs):
        super().__init__(**kwargs)
        if default_value:
            self.default_value = default_value
        else:
            self.default_value = Ratios(size=size, decimals=decimals)

    def validate(self, obj, value):
        if isinstance(value, Ratios):
            return value
        self.error(obj, value)

# Cell
class RatioLandingModel(Model):

    '''landing'''
    rowMax = Int(6)
    colMax = Int(6)
    n_row = Int()
    n_col = Int()
    rowName = Unicode()
    colName = Unicode()
    rowLevels = List()
    colLevels = List()
    rowLevelNames = List()
    colLevelNames = List()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLanding()

    ''' landing '''
    def setLanding(self):
        with self.hold_trait_notifications():
            self.n_row = 6
            self.n_col = 6
            self.rowName = 'Row'
            self.colName = 'Col'
            self.rowLevels = [str(i+1) for i in range(6)]
            self.colLevels = [str(i+1) for i in range(6)]

    @observe('n_row')
    def observe_n_row(self, change):
        self.rowLevelNames = [self.rowName + self.rowLevels[i] for i in range(change['new'])]

    @observe('n_col')
    def observe_n_col(self, change):
        self.colLevelNames = [self.colName + self.colLevels[i] for i in range(change['new'])]

    @observe('rowName')
    def observeRowName(self, change):
        self.rowLevelNames = [change['new'] + self.rowLevels[i] for i in range(self.n_row)]

    @observe('colName')
    def observeColName(self, change):
        self.colLevelNames = [change['new'] + self.colLevels[i] for i in range(self.n_col)]

    @observe('rowLevels')
    def observeRowLevels(self, change):
        self.rowLevelNames = [self.rowName + change['new'][i] for i in range(self.n_row)]

    @observe('colLevels')
    def observeColLevels(self, change):
        self.colLevelNames = [self.colName + change['new'][i] for i in range(self.n_col)]

    def changeRowLevel(self, i, value):
        self.rowLevels[i] = value
        if self.n > i:
            self.rowLevelNames[i] = self.name + value

# Cell
class RatioEstimatesModel(RatioLandingModel):

    decimals = Int(3)
    rowRatios = RatioArray()
    colRatios = RatioArray()
    total = Int()
    N = Int()
    colors = List()
    ns_cells = Array(dtype=np.dtype(int))
    percents = Array(dtype=np.dtype(float))
    rowTotals = Array(dtype=np.dtype(int))
    colTotals = Array(dtype=np.dtype(int))


    def setEstimates(self):
        with self.hold_trait_notifications():
            self.rowRatios = Ratios(size=self.n_row, decimals=self.decimals)
            self.colRatios = Ratios(size=self.n_col, decimals=self.decimals)
            self.colors = [['none'] * self.n_col for i in range(self.n_row)]
            self.total = self.n_row * self.n_col * 10
            percents = np.outer(self.rowRatios, self.colRatios)
            self.percents = np.around(percents, self.decimals)
            ns_cells = self.percents * self.total
            self.ns_cells = ns_cells.astype(np.dtype(int))
            self.rowTotals = np.sum(self.ns_cells, axis=1)
            self.colTotals = np.sum(self.ns_cells, axis=0)
            self.N = int(np.sum(self.ns_cells))

    @observe('total')
    def observeTotal(self, change):
        self.total = change['new']
        self.onChange()

    def incRowRatio(self, i, diff):
        self.rowRatios[i] += diff
        self.onRatioChange()

    def incColRatio(self, i, diff):
        self.colRatios[i] += diff
        self.onRatioChange()

    def onRatioChange(self):
        percents = np.outer(self.rowRatios, self.colRatios)
        self.percents = np.around(percents, self.decimals)
        self.onChange()

    def onChange(self):
        ns_cells = self.percents * self.total
        self.ns_cells = ns_cells.astype(np.dtype(int))
        self.rowTotals = np.sum(self.ns_cells, axis=1)
        self.colTotals = np.sum(self.ns_cells, axis=0)
        self.N = int(np.sum(self.ns_cells))


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_widgets.ipynb (unless otherwise specified).

__all__ = ['Widget', 'Output', 'IntSlider', 'FloatSlider', 'IntProgress', 'BoundedIntText', 'BoundedFloatText',
           'IntText', 'FloatText', 'ToggleButtons', 'Checkbox', 'Dropdown', 'RadioButtons', 'Select', 'Text',
           'Textarea', 'Label', 'HTML', 'HTMLMath', 'Image', 'Button', 'ColorPicker', 'Layout', 'HBox', 'AlignedLabel',
           'IntLabel', 'Grid']

# Cell
from fastcore.test import test_eq
from collections import namedtuple
import traitlets
from traitlets import Type, Unicode, observe, validate, Any, HasTraits, directional_link, link, Tuple, List, Int
from IPython.display import display, HTML

# Cell
from .items import ItemList, ItemArray, Item

# Cell
import ipywidgets as ipyw

# Cell
class Widget(ipyw.Widget, Item):
    def __init__(self, name=None, **kwargs):
        if name:
            self.name = name
            self.add_class(name)
        super().__init__(**kwargs)

    def visible(self):
        self.layout.visibility = 'visible'

    def invisible(self):
        self.layout.visibility = 'hidden'

    def hide(self):
        self.layout.display = 'inline'

    def show(self):
        self.layout.display = 'none'

# Cell
class Output(Widget, ipyw.Output):
    pass

class IntSlider(Widget, ipyw.IntSlider):
    pass

class FloatSlider(Widget, ipyw.FloatSlider):
    pass

class IntProgress(Widget, ipyw.IntProgress):
    pass

class BoundedIntText(Widget, ipyw.BoundedIntText):
    pass

class BoundedFloatText(Widget, ipyw.BoundedFloatText):
    pass

class IntText(Widget, ipyw.IntText):
    pass

class FloatText(Widget, ipyw.FloatText):
    pass

class ToggleButtons(Widget, ipyw.ToggleButtons):
    pass

class Checkbox(Widget, ipyw.Checkbox):
    pass

class Dropdown(Widget, ipyw.Dropdown):
    pass

class RadioButtons(Widget, ipyw.RadioButtons):
    pass

class Select(Widget, ipyw.Select):
    pass

class Text(Widget, ipyw.Text):
    pass

class Textarea(Widget, ipyw.Textarea):
    pass

class Label(Widget, ipyw.Label):
    pass

class HTML(Widget, ipyw.HTML):
    pass

class HTMLMath(Widget, ipyw.HTMLMath):
    pass

class Image(Widget, ipyw.Image):
    pass

class Button(Widget, ipyw.Button):
    pass

class ColorPicker(Widget, ipyw.ColorPicker):
    pass

class Layout(Widget, ipyw.Layout):
    pass

class HBox(Widget, ipyw.HBox):
    pass

# Cell
class AlignedLabel(Widget, ipyw.VBox):

    value = Unicode()

    def __init__(self, value=None, alignment='center', **kwargs):
        super().__init__(**kwargs)
        if value:
            self.value = value
        if alignment == 'left':
            justify_content = 'flex-start'
        elif alignment == 'right':
            justify_content = 'flex-end'
        else:
            justify_content = 'center'

        # create and add label
        self.label = Label(str(value), layout=ipyw.Layout(display='flex', justify_content=justify_content))
        self.children = (self.label, )

        self.link_to_label()

    def link_to_label(self):
        traitlets.link((self, 'value'), (self.label, 'value'))

# Cell
class IntLabel(AlignedLabel):

    value = Int(0)
    string = Unicode('0')

    @observe('value')
    def _observe_value(self, change):
        self.string = str(change['new'])
        return change['new']

    def link_to_label(self):
        traitlets.link((self, 'string'), (self.label, 'value'))

# Cell
class Grid(ipyw.GridspecLayout, Widget):

    def __init__(self, n_rows, n_columns, **kwargs):
        super().__init__(n_rows, n_columns, **kwargs)
        self._collections = {}

    def __getitem__(self, key):
        if isinstance(key, str):
            return self._collections[key]
        else:
            return super().__getitem__(key)

    def add_widget_array(self, widget_array, i=0, j=0, name=None):
        if name:
            widget_list.name = name
        for l in range(widget_array.n_rows):
            for m in range(widget_array.n_columns):
                self[i+l, j+m] = widget_array[l, m]

    def add_widget_list(self, widget_list, i=0, j=0, direction='row', name=None):
        if name:
            widget_list.name = name
        if hasattr(widget_list, '_size'):
            size = widget_list._size
        else:
            size = len(widget_list)
        for l in range(size):
            if direction == 'col':
                self[i+l, j] = widget_list[l]
            elif direction == 'row':
                self[i, j+l] = widget_list[l]


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/views.corr.ipynb (unless otherwise specified).

__all__ = ['CramerVLanding', 'CramerVEstimates', 'CramerV', 'TetraLanding', 'Tetra', 'CorrLanding',
           'SpearmansRhoLanding', 'PBiLanding', 'PearsonLanding', 'CorrView', 'Corr', 'SpearmansRho', 'PBi', 'Pearson']

# Cell
import ipywidgets as ipyw
import traitlets
from traitlets import observe, Unicode
import numpy as np
from fastcore.test import test_eq
from IPython.display import display, Javascript

# Cell
from ..widgets import Dropdown, BoundedFloatText, BoundedIntText, Textarea, Button
from ..widget_containers import VBox, Box, HBox
from ..models.corr import CramerVModel, TetraModel, SpearmansRhoModel, PBiModel, PearsonModel
from .ratio import RatioLanding, RatioEstimates
from ..mvc import Controller
import superpower_gui.superpower as sp

# Cell
class CramerVLanding(RatioLanding):
    pass

# Cell
class CramerVEstimates(HBox):

    def __init__(self, **kwargs):
        super().__init__(layout = ipyw.Layout(width='100%'), **kwargs)

        left = VBox(name='left', layout=ipyw.Layout(min_width='250px', width='30%'))
        self.right = VBox(name='right', layout=ipyw.Layout(width='auto'))

        ''' left '''
        self.textarea = Textarea(layout=ipyw.Layout(width="94%", height="150px"), disabled=True)
        self.effectButton = Button(description='Show effect sizes', button_style='success',
            layout=ipyw.Layout(min_width='220px', max_width='220px',margin = '2px 2px 2px 2px'))
        left.children = (self.textarea, self.effectButton)

        self.children = (left, self.right)

# Cell
class CramerV(CramerVEstimates, Controller):

    def __init__(self, model, **kwargs):
        super().__init__(**kwargs)
        self.model = model

        self.effectSizes = []
        self.ratio = RatioEstimates(model)
        self.rowCompareButton = Button(layout=ipyw.Layout(width='90%', height='auto'))
        self.ratio[0,0:2] = self.rowCompareButton
        self.right.children += (self.ratio, )

        self.initialize()

        ''' observes '''
        self.rowCompareButton.on_click(self.observeRowCompare)
        self.ratio.colorIntArray.observe(self.observeColors, names='color', type='child_change')
        self.effectButton.on_click(self.observeEffectButton)

    def initialize(self):
        self.ratio.initialize()
        self.textarea.value = ''
        self.observeRowCompare(None)
        if not self.model.hasColors:
            self.rowCompareButton.disabled = True
            self.effectButton.disabled = True

    def setEstimates(self):
        self.ratio.setEstimates()
        super().setEstimates()

    def observeRowCompare(self, button):
        if self.model.rcomp:
            self.model.rcomp = False
            self.rowCompareButton.description = 'Column Compare'
        else:
            self.model.rcomp = True
            self.rowCompareButton.description = 'Row Compare'

    def observeColors(self, change):
        i, j = change['index']
        self.model.changeColor(i, j, change['new'])
        if self.model.hasColors:
            self.effectButton.disabled = False
            self.rowCompareButton.disabled = False
        else:
            self.effectButton.disabled = True
            self.rowCompareButton.disabled = True

    def observeEffectButton(self, button):
        if self.model.hasColors:
            try:
                res = self.model.runEffectSize()
                for i, string in enumerate(self.model.strings):
                    if i == len(self.model.strings) - 1:
                        self.textarea.value += string + ' = %.3f\n\n' % res['df'].power[0]
                    else:
                        self.textarea.value += string + ':'
                self.effectSizes.append(res)
            except Exception as e:
                display(Javascript('alert("Error: %s\\n\\nDo you mean to be comparing rows or columns?")' % str(e)))

    def run(self):
        result = self.model.run()
        if result:
            result['effectSizes'] = self.effectSizes
        return result

# Cell
class TetraLanding(RatioLanding):
    pass

class Tetra(CramerV):
    pass

# Cell
class CorrLanding(Box):

    def __init__(self, model, **kwargs):
        super().__init__(**kwargs)
        self.model = model

    def setLanding(self):
        pass

class SpearmansRhoLanding(CorrLanding):
    pass

class PBiLanding(CorrLanding):
    pass

class PearsonLanding(CorrLanding):
    pass

# Cell
class CorrView(VBox):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.function = SpearmansRhoModel()
        self.r = BoundedFloatText(description='r', min=0, max=1, step=0.01)
        self.N = BoundedIntText(description='N', min=0)
        self.children = (self.N, self.r)

# Cell
class Corr(CorrView):

    def __init__(self, model, **kwargs):
        super().__init__(**kwargs)
        self.model = model
        traitlets.link((self.model, 'r'), (self.r, 'value'))
        traitlets.link((self.model, 'N'), (self.N, 'value'))

    def run(self):
        return self.model.run(**self.model.getTraits())

# Cell
class SpearmansRho(Corr):
    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)

class PBi(Corr):
    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)

class Pearson(Corr):
    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/superpower.corr.ipynb (unless otherwise specified).

__all__ = ['CorrPearson', 'CorrPBi', 'CorrSpearmansRho', 'CorrCramerV', 'CorrTetra']

# Cell
import numpy as np

# Cell
from ..log import HANDLERS
from ..rpy import RFunction, R_CONSOLE

# Cell
class CorrPearson(RFunction): # one
    argNames = ['N', 'r', 'alternative', 'pVal']
    def __init__(self, argNames=argNames, **kwargs):
        super().__init__(name='corrPearson', argNames=argNames, **kwargs)

class CorrPBi(RFunction): # two
    argNames = ['N', 'r', 'tail', 'pVal']
    def __init__(self, argNames=argNames, **kwargs):
        super().__init__(name='corrPBi', argNames=argNames, **kwargs)

class CorrSpearmansRho(RFunction): # one
    argNames = ['N', 'r', 'alternative', 'pVal']
    def __init__(self, argNames=argNames, **kwargs):
        super().__init__(name='corrSpearmansRho', argNames=argNames, **kwargs)

class CorrCramerV(RFunction): # one
    argNames=['n_row', 'n_col', 'alternative', 'ns_cells', 'N', 'pVal']
    def __init__(self, argNames=argNames, **kwargs):
        super().__init__(name='corrCramerV', argNames=argNames, **kwargs)

class CorrTetra(RFunction): # two
    argNames=['n_row', 'n_col', 'tail', 'ns_cells', 'N', 'pVal']
    def __init__(self, argNames=argNames, **kwargs):
        super().__init__(name='corrTetra', argNames=argNames, **kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/results.ipynb (unless otherwise specified).

__all__ = ['rpy2', 'logger', 'ResultsModel', 'ResultsView', 'Results']

# Cell
import pandas as pd
import numpy as np
from traitlets import List, HasTraits

# Cell
from .widget_containers import VBox
from .widgets import Output, Select
from ipywidgets import HBox, Layout
from IPython.core.display import display, clear_output, HTML
from .config import CONFIG, APP_DIR
from .log import HANDLERS
from .rpy import RFunction, R_CONSOLE
rpy2 = R_CONSOLE.rpy2
logger = HANDLERS.getLogger(__name__)

# Cell
class ResultsModel(HasTraits):

    data = List([])

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def addResult(self, result):
        self.data += [dict(result)] #because results is a UserDict


# Cell
# View
class ResultsView(VBox):

    HEADER  = ['ID             ','Name           ','Date                  ','Runtime   ','Input                                    ']

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        header = self.columnize(self.HEADER,[self.HEADER])
        header = Select(options=header,disabled=True,value=None,layout=Layout(height='2em',width='99%'))
        header.add_class('selmono') # See custom CSS

        # Fixed width results selection widet
        self.select = Select(rows=10,options=[],value=None,layout={'width':'99%'})
        self.select.add_class('selmono') # See custom CSS

        self.output = Output()

        #self.style  = HTML('<style> .selmono { font-family: "Lucida Console", Monaco, monospace !important;} </style>')
        #self.children += (self.style, header,self.select,self.output, )
        self.children += (header,self.select,self.output, )

    def columnize(self,guide,data,sep=' '):
        '''Pad or truncate columns in text lines based on guide'''
        rows = []

        for r in range(len(data)):
            row = ''

            for c in range(len(data[r])): # guide)):
                text      = data[r][c]
                col_width = len(guide[c])
                diff      = col_width - len(data[r][c])

                if diff < 0:
                    row += text[:diff]         + sep
                else:
                    row += text + (' ' * diff) + sep

            rows.append(row)

        return rows

# Cell
# Controller
class Results(ResultsView):

    VALUE = 'value' # for observe calls

    def __init__(self, model, **kwargs):
        super().__init__(**kwargs)

        self.model = model
        self.select.observe(self.onSelect,self.VALUE)

    def addResult(self, result):
        self.model.addResult(result)

        # Add to select widget: temporarily suspend callback, update result select w/new data
        self.select.unobserve(self.onSelect,self.VALUE)
        item = [str(result['id']),result['name'],result['date'],
                str(result['runtime']),str(result['kwargs'])]
        item = self.columnize(self.HEADER,[item])[0]  # TODO adjust columnize() to do one-at-a-time
        newOptions = list(self.select.options)
        newOptions.append(tuple([item,result]))
        self.select.options = newOptions
        self.select.value = None
        self.select.observe(self.onSelect,self.VALUE)

    def onSelect(self,change):
        with self.output:
            clear_output(wait=True)
            display(change['owner'].value['df'])



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/controller.ipynb (unless otherwise specified).

__all__ = ['Selection', 'Controller']

# Cell
from IPython.display import display, Javascript
import json
import traitlets
from traitlets import HasTraits, Int, Unicode, observe
import ipywidgets as ipyw

# Cell
from .widgets import IntSlider, Button, Dropdown, FloatText
from .widget_containers import Tab, SelectionVBox, VBox, Box, Accordion
from .views.tabs import LandingTab, EstimatesTab, ResultsTab, SettingsTab
from .log import HANDLERS
from .models.chisq import *
from .views.chisq import *
from .models.corr import *
from .views.corr import *
from .results import ResultsModel, Results
from .settings import SettingsModel, Settings
from .rpy import R_CONSOLE

# Cell
from collections import namedtuple
Selection = namedtuple('Selection', ['label', 'LandingModel', 'Model', 'Landing', 'Estimates'])

# Cell
class Controller(traitlets.HasTraits):

    modelName = Unicode()
    corrModelName = Unicode()

    def __init__(self):

        ''' log handlers'''
        self.logger = HANDLERS.getLogger(__name__)
        self.logger.info('Autosave off.')
        display(Javascript('Jupyter.notebook.set_autosave_interval(0);'))
        self.handlerWidget = HANDLERS.handlers['widget']
        HANDLERS.loadCSS()

        ''' create views and models '''
        self.create_global_widgets()
        self.define_box_selections()
        self.initialize_mvcs()
        self.view = self.tab_based_view()
        self.add_observations()

        ''' initialize '''
        self.corrDropdown.value = 'cramerv'
        self.dropdown.value = 'chisq'


    def _ipython_display_(self):
        display(self.view)

    def create_global_widgets(self):

        ''' landing '''
        self.corrDropdown = Dropdown(description='Coefficient: ',
            layout=ipyw.Layout(max_width='350px',description_width='100px'))
        self.corrSelection = SelectionVBox()
        self.corrBox = VBox(children=(self.corrDropdown, self.corrSelection))

        self.landingSelection = SelectionVBox()
        self.dropdownOptions = [('Correlation', 'corr')] # populated in self.initialize_view_and_models
        self.dropdown = Dropdown( options=self.dropdownOptions, description='Model Class: ',
                layout=ipyw.Layout(max_width='300px', description_width='80px'))

        self.okayButton = Button(description='Okay', button_style='success')

        ''' estimates '''
        self.estimatesBox = Box()
        self.runButton = Button(description='Run', button_style='success')
        self.restartButton = Button(description='Restart')

        ''' results '''
        self.repopulateLanding = Button(description='Repopulate Model Tab')
        self.repopulateEstimates = Button(description='Repopulate Estimates Tab')

        ''' settings '''
        self.logAccordion = Accordion(children=(self.handlerWidget, ))

    def define_box_selections(self):

        self.options = {}
        self.options['chisq'] = Selection('Chi Squared', ChiSqLandingModel, ChiSqModel, ChiSqLanding, ChiSq)

        self.corrOptions = {}
        self.corrOptions['tetra'] = Selection('Tetrachoric', TetraLandingModel, TetraModel, TetraLanding, Tetra)
        self.corrOptions['cramerv'] = Selection("Cramer's V", CramerVLandingModel, CramerVModel, CramerVLanding, CramerV)
        self.corrOptions['pears'] = Selection("Pearson's Correlation Coefficient", CorrLandingModel, PearsonModel, PearsonLanding, Pearson)
        self.corrOptions['spear'] = Selection("Spearman's Rho", CorrLandingModel, SpearmansRhoModel, SpearmansRhoLanding, SpearmansRho)
        self.corrOptions['point'] = Selection("Point Biserial", CorrLandingModel, PBiModel, PBiLanding, PBi)

    def initialize_single_mvc(self, options, selection, dropdown):
        for name, option in options.items():
            label, LandingModel, Model, Landing, Estimates = option
            landingModel = LandingModel()
            self.landingModels[name] = landingModel
            self.Model[name] = Model
            dropdown.options += ((label, name), )
            selection.add_child(Landing(landingModel), name)
            self.Estimates[name] = Estimates

    def initialize_mvcs(self):

        ''' selectable boxes in landing '''
        self.Model = {}
        self.Estimates = {}
        self.landingModels = {}
        self.initialize_single_mvc(self.options, self.landingSelection, self.dropdown)

        self.corrLandingModels = {}
        self.initialize_single_mvc(self.corrOptions, self.corrSelection, self.corrDropdown)
        self.landingSelection.add_child(self.corrBox, 'corr')

        '''other mvcs'''
        self.resultsModel = ResultsModel()
        self.settingsModel = SettingsModel()

        self.resultsView = Results(self.resultsModel)
        self.settingsView = Settings(self.settingsModel)

    def tab_based_view(self):
        landingTab = LandingTab(self.dropdown, self.landingSelection, self.okayButton, name='Model')
        estimatesTab = EstimatesTab(self.estimatesBox, self.runButton, name='Estimates')
        resultsTab = ResultsTab(self.restartButton, self.repopulateLanding, self.repopulateEstimates, self.resultsView, name='Results')
        settingsTab = SettingsTab(self.settingsView, self.logAccordion, name='Settings')
        return Tab(children=(landingTab, estimatesTab, resultsTab, settingsTab))

    def add_observations(self):
        self.okayButton.on_click(self.on_okayButton_click)
        self.runButton.on_click(self.on_runButton_click)
        self.restartButton.on_click(self.on_restartButton_click)
        self.repopulateLanding.on_click(self.on_repopulateLanding_click)
        self.repopulateEstimates.on_click(self.on_repopulateEstimates_click)

        self.dropdown.observe(self.on_dropdown_change, names='value')
        self.corrDropdown.observe(self.on_corrDropdown_change, names='value')

    def on_dropdown_change(self, change):
        if change['new'] == 'corr':
            self.modelName = self.corrDropdown.value
            self.landingSelection.selected_name = 'corr'
        else:
            self.modelName = change['new']
            self.landingSelection.selected_name = self.modelName

    def on_corrDropdown_change(self, change):
        self.corrSelection.selected_name = change['new']
        self.modelName = change['new']

    ''' on button click'''
    def on_okayButton_click(self, button):
        model = self.Model[self.modelName](**self.landingModels[self.modelName].getTraits())
        model.setEstimates()
        self.estimates = self.Estimates[self.modelName](model, name=self.modelName)
        self.estimatesBox.children = (self.estimates, )
        self.view.selected_name = 'Estimates'

    def on_runButton_click(self, button):
        self.update_model_with_settings()
        result = self.estimates.run()
        if result:
            self.resultsView.addResult(result)
            self.view.selected_name='Results'

    def on_restartButton_click(self, button):
        if self.modelName == 'corr':
            self.corrSelection[self.corrModelName].setLanding()
        else:
            self.landingSelection[self.modelName].setLanding()
        self.view.selected_name= 'Model'

    def on_repopulateLanding_click(self, button):
        self.view.selected_name = 'Model'

    def on_repopulateEstimates_click(self, button):
        self.view.selected_name = 'Estimates'

    def update_model_with_settings(self):
        if self.modelName in ('chisq','tetra', 'cramerv'):
            self.estimates.model.pVal = self.settingsModel.alpha
            self.estimates.model.tail = self.settingsModel.tail
        else:
            self.estimates.model.setTraits(**self.settingsModel.getTraits())


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/models.corr.ipynb (unless otherwise specified).

__all__ = ['CramerVLandingModel', 'AddStringsMixin', 'CramerVModel', 'TetraLandingModel', 'TetraModel',
           'CorrLandingModel', 'CorrModel', 'SpearmansRhoModel', 'PearsonModel', 'PBiModel']

# Cell
import ipywidgets as ipyw
import traitlets
from traitlets import observe, Unicode, Bool, Int, Float
from traittypes import Array
import numpy as np
from fastcore.test import test_eq
from IPython.display import display, Javascript

# Cell
from .ratio import RatioLandingModel, RatioEstimatesModel
from ..mvc import Model
from ..items import Collection
from ..traits import LinkAllMixin
from ..rpy import RMixins
from ..superpower.corr import CorrCramerV, CorrTetra, CorrSpearmansRho, CorrPBi, CorrPearson

# Cell
class CramerVLandingModel(RatioLandingModel):
    pass

# Cell
class AddStringsMixin():
    ''' Like numpy.char.add but it works on dtype np.object'''

    def add_strings(self, first, second):
        if isinstance(first,str):
            first = [first] * len(second)
        if isinstance(second,str):
            second = [second] * len(first)
        if len(first) != len(second):
            raise ValueError('If adding arrays, they must have matching lengths')
        return np.array([[first[i] + second[j] for j in range(len(second))] for i in range(len(first))], dtype=np.object)

# Cell
class CramerVModel(RatioEstimatesModel, AddStringsMixin):

    pVal = Float()
    rcomp = Bool()
    mata = Array(dtype=np.dtype(int))
    matb = Array(dtype=np.dtype(int))
    matrix = Array(dtype=np.dtype(int))
    w = Float()
    hasColors = Bool()
    tail = Unicode()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.cramerv = CorrCramerV()

    def setEstimates(self):
        super().setEstimates() #ratio model
        self.pVal = 0.05
        self.w = 0.3
        self.rcomp = False
        self.hasColors = False
        self.tail = 'two'

    def changeColor(self, i, j, color):
        self.colors[i][j] = color
        if np.all(np.array(self.colors) == 'none'):
            self.hasColors = False
        else:
            self.hasColors = True

    def runEffectSize(self):
        if self.hasColors:
            self.validateColors()
            kwargs = self.getTraits(['matrix', 'pVal', 'tail'])
            matrix = kwargs.pop('matrix')
            kwargs['n_row'], kwargs['n_col'] = matrix.shape
            kwargs['ns_cells'] = matrix.flatten()
            kwargs['alternative'] = kwargs.pop('tail')
            if kwargs['alternative'] == 'two':
                kwargs['alternative'] = 'two sided'
            return self.cramerv.run(**kwargs)

    def run(self):
        if self.hasColors:
            return self.runEffectSize()
        else:
            kwargs = self.getTraits(['ns_cells', 'pVal', 'tail'])
            kwargs['n_row'], kwargs['n_col'] = kwargs['ns_cells'].shape
            kwargs['ns_cells'] = kwargs['ns_cells'].flatten()
            kwargs['alternative'] = kwargs.pop('tail')
            if kwargs['alternative'] == 'two':
                kwargs['alternative'] = 'two sided'
            return self.cramerv.run(**kwargs)

    def validateColors(self):
        self.mata = []
        self.matb = []
        self.strings = []
        colors = np.array(self.colors)
        ns_cells = np.array(self.ns_cells)
        strings = self.add_strings(self.rowLevelNames, self.colLevelNames)
        first = None
        switched = False
        done = False
        i = 0
        if self.rcomp:
            length = self.n_col
            stack = np.vstack
        else:
            length = self.n_row
            stack = np.column_stack
        while (i < length):
            if self.rcomp:
                color_vec = colors[i]
                vector = ns_cells[i]
                string = strings[i]
            else:
                color_vec = colors[:,i]
                vector = ns_cells[:,i]
                string = strings[:,i]
            if not first:
                red = color_vec == 'red'
                green = color_vec == 'green'
                if np.any(red) or np.any(green):
                    if np.any(red):
                        first = 'red'
                        second = 'green'
                        self.pattern = red
                    elif np.any(green):
                        first = 'green'
                        second = 'red'
                        self.pattern = green
                    self.mata = vector[self.pattern]
                    self.strings.append(np.array2string(string[self.pattern], separator=',', max_line_width=10000, formatter={'object':lambda x: x}))
                    if not self.rcomp:
                        self.mata = np.transpose([self.mata])
            elif not switched: # if we haven't switched to the second color yet
                vec = color_vec == first
                if np.any(vec): # if this is still a vector with the first color
                    assert np.array_equal(vec, self.pattern), 'Selected self.colors should be the same for every selected %s' % string
                    self.mata = stack((self.mata, vector[self.pattern]))
                    self.strings.append(np.array2string(string[self.pattern], separator=',', max_line_width=10000, formatter={'object':lambda x: x}))
                else: # if this vector is no longer the first color
                    vec = color_vec == second
                    if not np.any(vec):
                        raise Exception('Both colors must be present.')
                    else:
                        assert np.array_equal(vec, self.pattern), 'Selected self.colors should be the same for every selected %s' % string
                        self.matb = vector[self.pattern]
                        self.strings.append(np.array2string(string[self.pattern], separator=',', max_line_width=10000, formatter={'object':lambda x: x}))
                        if not self.rcomp:
                            self.matb = np.transpose([self.matb])
                        switched = True
            elif not done:
                vec = color_vec == second
                if np.any(vec):
                    assert np.array_equal(vec, self.pattern), 'Selected self.colors should be the same for every selected %s' % string
                    self.matb = stack((self.matb, vector[self.pattern]))
                    self.strings.append(np.array2string(string[self.pattern], separator=',', max_line_width=10000, formatter={'object':lambda x: x}))
                else:
                    done = True
            else:
                if np.any(color_vec != 'none'):
                    raise Exception('Misplaced color selections.')
            i += 1
        self.matrix = stack((self.mata, self.matb))

# Cell
class TetraLandingModel(RatioLandingModel):
    pass

# Cell
class TetraModel(CramerVModel, RMixins):

    tail = Unicode()
    pVal = Float()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.function = CorrTetra()

    def setLanding(self):
        super().setLanding()
        self.tail = 'two'
        self.pVal = 0.05

    def run(self):
        #corrTetra(tail = "two", conttab = matrix(rep(10, 36), nrow = 6, byrow = TRUE), obs = TRUE, pVal = 0.05)
        kwargs = self.getTraits(['ns_cells', 'pVal', 'tail', 'n_row'])
        kwargs['conttab'] = self.matrix_func(kwargs.pop('ns_cells').flatten(), nrow = kwargs['n_row'], byrow = True)
        kwargs['obs'] = True
        if kwargs['tail'] != 'two':
            kwargs['tail'] = 'one'
        del kwargs['n_row']
        print(kwargs)
        return self.function.run(**kwargs)

# Cell
class CorrLandingModel(Model):

    def setLanding(self):
        pass

class CorrModel(CorrLandingModel):

    r = Float()
    N = Int()
    alpha = Float()
    tail = Unicode()

    def setEstimates(self):
        self.alpha = 0.05
        self.tail = 'two'
        self.r = 0.5
        self.N = 100

class SpearmansRhoModel(CorrModel):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.function = CorrSpearmansRho()

    def run(self, **kwargs):
        kwargs = self.getTraits()
        kwargs['alternative'] = kwargs.pop('tail')
        if kwargs['alternative'] == 'two':
            kwargs['alternative'] = 'two sided'
        kwargs['pVal'] = kwargs.pop('alpha')
        return self.function.run(**kwargs)

# Cell
class PearsonModel(CorrModel):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.function = CorrPearson()

    def run(self, **kwargs):
        kwargs = self.getTraits()
        kwargs['alternative'] = kwargs.pop('tail')
        if kwargs['alternative'] == 'two':
            kwargs['alternative'] = 'two sided'
        kwargs['pVal'] = kwargs.pop('alpha')
        return self.function.run(**kwargs)

# Cell
class PBiModel(CorrModel):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.function = CorrPBi()

    def run(self, **kwargs):
        kwargs = self.getTraits()
        if kwargs['tail'] != 'two':
            kwargs['tail'] = 'one'
        kwargs['pVal'] = kwargs.pop('alpha')
        print(kwargs)
        return self.function.run(**kwargs)